﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using IDTechSDK;
using System.Threading;
using System.IO;
using System.Diagnostics;
using System.IO.Ports;
using System.Runtime.InteropServices;
using System.Resources;
using System.Net;
using System.Threading.Tasks;
using System.Security.Cryptography.X509Certificates;
using System.Net.Sockets;
using System.Net.NetworkInformation;
using System.Collections;
using System.Runtime.Remoting.Metadata.W3cXsd2001;
using System.Security.Cryptography;
using Microsoft.VisualBasic;
using WorldpayUtility.Properties;
using System.Windows;
using System.Reflection;
using IDTechSDK.Configs;

namespace WorldpayUtility
{
    public partial class Worldpay_Utility : Form
    {
        delegate void SetTextCallback(string text, string ident);
        delegate void SetTitleTextCallback(string text);

        delegate void SetTextCallbackWithTextAlign(string text, string ident, bool withAlign = false);

        delegate void DisplayMSRPrompt(string text, string ident);

        delegate void refreshAllDevices();

        Dictionary<string, Config> config = new Dictionary<string, Config>();
        bool inStartup = false;
        public Worldpay_Utility()
        {
            inStartup = true;
            InitializeComponent();
            //Set up main callback used by all devices
            IDT_Device.setCallbackIP(MessageCallBack);
            inLoad = true;
            //acceptorID.Text= Settings.Default.acceptorid ;
            //accountToken.Text=Settings.Default.accounttoken;
            //accountID.Text = Settings.Default.accountid;
            //terminalid.Text = Settings.Default.terminalid;
            inLoad = false;

            string ver = IDT_Device.SDK_Version();
            string ver2 = Application.ProductVersion;
            inStartup = false;
            SetTitleText("ID Tech WorldPay Utility: " + "SDK: " + ver + "  /  App: " + ver2);
            IDT_Device.startUSBMonitoring();
            Profile.discoverDevice();



        }


        private string _lastIP = "";
        private string _lastFW = "";
        private string _lastTransfer = "";
        private string _lastCAPK = "";



        bool isVP8800()
        {
            string sel = (string)connectedDevicesList.SelectedItem;
            if (sel == null) sel = "";
            string sel2 = (string)connectedDevicesList.SelectedItem;
            if (sel2 == null) sel2 = "";
            return sel.Equals("VP8800") || sel2.Equals("VP8800") || sel.Equals(NEONAME) || sel2.Equals("NEO2");
        }

        private void closeAllDevices()
        {
            Profile.closeAllCommConnections();
        }

        private void UniPayDemo_FormClosing(object sender, FormClosingEventArgs e)
        {

            closeAllDevices();
        }

        public static String GetTimestamp()
        {
            DateTime value = DateTime.Now;
            return (value.ToString("HH:mm:ss.fff") + "  ");
        }



        private void refreshDevices()
        {
            if (connectedDevicesList.InvokeRequired)
            {
                refreshAllDevices d = new refreshAllDevices(_refreshDevices);
                Invoke(d, new object[] { });
            }
            else
            {
                _refreshDevices();
            }
            //buildTreeView();
        }

        private string NEONAME = "NEO2";
        private int usbCount = 0;



        private void _refreshDevices()
        {

            
            List<string> ulist = Profile.getAllConnections();
           

            string sel = "";
            bool usameDevice = false;





            if (connectedDevicesList != null && connectedDevicesList.SelectedItem != null)
            {
                sel = (string)connectedDevicesList.SelectedItem;
            }

            bool reloadList = false;
            if (ulist.Count != usbCount) reloadList = true;
            if (ulist.Count == usbCount && sel != null)
            {
                if (!ulist.Contains(sel)) reloadList = true;
            }
            if ((ulist.Count == usbCount) && ulist.Count > 0)
            {
                foreach (string str in connectedDevicesList.Items)
                {
                    if (!ulist.Contains(str)) reloadList = true;
                }
            }
            if (connectedDevicesList.SelectedItem == null)
            {
                reloadList = true;
            }
            if (connectedDevicesList.SelectedItem != null && connectedDevicesList.SelectedItem.ToString().Length == 0)
            {
                reloadList = true;
            }
            if (reloadList)
            {
                usbCount = ulist.Count;


                connectedDevicesList.Items.Clear();



                if (ulist != null && ulist.Count > 0)
                {
                    foreach (string item in ulist)
                    {
                        connectedDevicesList.Items.Add(item);
                        if (sel.Length > 0 && sel.Equals(item))
                        {
                            usameDevice = true;
                        }
                    }
                }
            }


            if (connectedDevicesList.Items.Count > 0)
            {
                if (usameDevice)
                { connectedDevicesList.SelectedItem = sel; }
                else if (connectedDevicesList.SelectedItem == null) { 
                    connectedDevicesList.SelectedIndex = 0; 
                }
            }



            List<string> tabPages = new List<string>();
            foreach (TabPage page in Devices.TabPages) tabPages.Add(page.Text);

            if (connectedDevicesList.Items.Count == 0)
            {
                if (tabPages.Count == 1 && tabPages[0].Equals("No Device")) return;
                foreach (string str in tabPages) removeTab(str);
                return;
            }
            foreach (string str in connectedDevicesList.Items)
            {
                if (tabPages.Contains(str))
                {
                    tabPages.Remove(str);
                }
                else
                {
                    
                        
                        getLogTab(str);
  

                    
                }
            }
            foreach (string str in tabPages) removeTab(str);
            if (connectedDevicesList != null && connectedDevicesList.SelectedItem != null)
            {
                sel = (string)connectedDevicesList.SelectedItem;
            }
            TabPage pg = getLogTab(sel);
            if (pg != null) Devices.SelectedTab = pg;

        }



        private void vcComplete(RETURN_CODE verify, string ident)
        {
            IDTechComm comm = Profile.getComm(ident);
            if (verify == RETURN_CODE.RETURN_CODE_SDK_VIVOCONFIG_VERIFY_SUCCESS)
            {
                MessageBox.Show("Verification process completed successfully. No differences found.", "Verify Success", MessageBoxButtons.OK, MessageBoxIcon.Information);
                comm.verifyWarning = false;

            }
            if (verify == RETURN_CODE.RETURN_CODE_SDK_VIVOCONFIG_VERIFY_SUCCESS_WITH_WARNING)
            {
                MessageBox.Show("Verification process completed with warnings. Please evaluate the log.", "Verify with Errors", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                comm.verifyWarning = false;

            }
            if (verify == RETURN_CODE.RETURN_CODE_SDK_VIVOCONFIG_VERIFY_FAILED)
            {
                MessageBox.Show("Verification process FAILED. Please evaluate the log.", "Verify Failed", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                comm.verifyWarning = false;

            }

            if (verify == RETURN_CODE.RETURN_CODE_SDK_VIVOCONFIG_WRITE_SUCCESS)
            {
                MessageBox.Show("Write process completed successfully.", "Write Success", MessageBoxButtons.OK, MessageBoxIcon.Information);
                comm.verifyWarning = false;

            }
            if (verify == RETURN_CODE.RETURN_CODE_SDK_VIVOCONFIG_WRITE_SUCCESS_WITH_WARNING)
            {
                //Ignoring errors because universal config file may contain tags not availalbe to all devices, so warnings are expected.
                //MessageBox.Show("Write process completed with warnings. Please evaluate the log.", "Write with Errors", MessageBoxButtons.OK, MessageBoxIcon.Warning);

                MessageBox.Show("Write process completed successfully.", "Write Success", MessageBoxButtons.OK, MessageBoxIcon.Information);
                comm.verifyWarning = false;

            }
            if (verify == RETURN_CODE.RETURN_CODE_SDK_VIVOCONFIG_WRITE_FAILED)
            {
                MessageBox.Show("Write process FAILED. Please evaluate the log.", "Write Failed", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                comm.verifyWarning = false;

            }
 
    
        }



        private void MessageCallBack(IDTechSDK.IDT_DEVICE_Types type, DeviceState state, byte[] data,
            IDTTransactionData cardData, EMV_Callback emvCallback, RETURN_CODE transactionResultCode, string ident)
        {
            
            if (InvokeRequired)
            {
                try
                {

                    Invoke((MethodInvoker)delegate
                    {
                        MessageCallBack(type, state, data, cardData, emvCallback, transactionResultCode, ident);
                    });

                    return;
                }
                catch
                {

                }

            }



            IDTechComm comm = Profile.getComm(type, DEVICE_INTERFACE_Types.DEVICE_INTERFACE_USB, ident);
            if (comm == null) comm = Profile.getComm(type, DEVICE_INTERFACE_Types.DEVICE_INTERFACE_SERIAL, ident);
            if (comm == null) comm = Profile.getComm(type, DEVICE_INTERFACE_Types.DEVICE_INTERFACE_BLE, ident);
            if (comm == null) comm = Profile.getComm(type, DEVICE_INTERFACE_Types.DEVICE_INTERFACE_AUDIO_JACK, ident);
            if (comm == null) comm = Profile.getComm(type, DEVICE_INTERFACE_Types.DEVICE_INTERFACE_IP, ident);
            DEVICE_INTERFACE_Types connect = DEVICE_INTERFACE_Types.DEVICE_INTERFACE_UNKNOWN;
            if (comm != null)
            {
                connect = comm.getDeviceConnection();
            }
             
  
            if (comm != null)
            {
                if (comm.isKB && state == DeviceState.Connected)
                {
                    MessageBox.Show("KB-Mode device detected.  This application is for HID-Mode devices only. Please put your device into HID-Mode and try again","KB Mode Device Detected",MessageBoxButtons.OK, MessageBoxIcon.Error);
                    Profile.closeDevice(comm.ident);
                }
                if (comm.isKB) return;
            }
            switch (state)
            {

  
                case DeviceState.DefaultDeviceTypeChange:
                    // refreshDevices();
                    SetOutputText("\nSDK Default Device =  " + IDTechSDK.Profile.IDT_DEVICE_String(type, connect));
                   // SetTitleText("SDK Demo:  " + IDTechSDK.Profile.IDT_DEVICE_String(type, connect, ident),ident);
                   // refreshDevices();
                

                    break;
                case DeviceState.Connected:
                        SetOutputText(GetTimestamp() + "\nConnected " +
                IDTechSDK.Profile.IDT_DEVICE_String(type, connect, ident));
                       // SetTitleText(IDTechSDK.Profile.IDT_DEVICE_String(type, connect, ident), ident);
                        refreshDevices();
                        if (sel == null || sel.Length == 0)
                        {
                            if (ident == null) ident = "";
                            sel = ident;
                        }
                    
  
                    break;
                case DeviceState.Disconnected:
                    if (sel != null)
                    {
                        if (ident == null) ident = "";
                        if (sel.Equals(ident)) sel = "";
                    }
                
                    SetOutputText(GetTimestamp() + "\nDisconnected " +
                                  IDTechSDK.Profile.IDT_DEVICE_String(type, connect, ident));
                    //removeTab(ident);
                    refreshDevices();
                    break;
                case DeviceState.ConnectionFailed:
                    SetOutputText("\nConnection Failed\n");
                    break;
                case DeviceState.RKI:                 
                    break;
                             
                case DeviceState.SocketEstablished:
                    SetOutputText(" Socket connection established to: " + ident + " \r\n");
                    break;
                case DeviceState.SocketFailed:
                    SetOutputText(" Socket connection failed to: " + ident +  " \r\n");
                    if (data != null)
                    {
                        string str = ASCIIEncoding.ASCII.GetString(data);
                        SetOutputText(str + " \r\n", ident);
                    }

                    break;
                case DeviceState.ViVOconfig:
                    string outputtext = Encoding.UTF8.GetString(data, 0, data.Length);
                    SetOutputText(outputtext, ident);
                    vcComplete(transactionResultCode, ident);
                    break;
                case DeviceState.TransactionData:


                    if (cardData == null)
                    {
                        SetOutputText("Card Data Null - Breaking", ident);
                        break;
                    }

                    SetOutputText("cardData.emv_resultCode = " + cardData.emv_resultCode.ToString(), ident);
   


                    if (cardData.Event == EVENT_TRANSACTION_DATA_Types.EVENT_TRANSACTION_PIN_DATA ||
                        cardData.Event == EVENT_TRANSACTION_DATA_Types.EVENT_TRANSACTION_DATA_CARD_DATA ||
                        cardData.Event == EVENT_TRANSACTION_DATA_Types.EVENT_TRANSACTION_DATA_EMV_DATA)
    
                    {
                           
                        displayCardData(cardData, ident);        
                    }

                    break;

                case DeviceState.EMVCallback:


                    if (emvCallback == null) break;
                    if (emvCallback.callbackType == EMV_CALLBACK_TYPE.EMV_CALLBACK_TYPE_LCD) //LCD Callback Type
                    {
                        if (emvCallback.lcd_displayMode == EMV_LCD_DISPLAY_MODE.EMV_LCD_DISPLAY_MODE_CLEAR_SCREEN)
                        {

                            return;
                        }
                        else if (emvCallback.lcd_displayMode == EMV_LCD_DISPLAY_MODE.EMV_LCD_DISPLAY_MODE_MESSAGE || emvCallback.lcd_displayMode == EMV_LCD_DISPLAY_MODE.EMV_LCD_DISPLAY_MODE_AR_MESSAGE2 || emvCallback.lcd_displayMode == EMV_LCD_DISPLAY_MODE.EMV_LCD_DISPLAY_MODE_AR_MESSAGE3)
                        {
                            showMessageOnLCD(emvCallback.language, emvCallback.lcd_messages, ident);
                        }
                        else if (emvCallback.lcd_displayMode == EMV_LCD_DISPLAY_MODE.EMV_LCD_DISPLAY_MODE_AR_MESSAGE4)
                        {
                            showMessageOnLCD(emvCallback.language, emvCallback.lcd_messages, ident);

                            tbOutputAppendText("Mode 4 Callback - Awaiting Response.  \r\n", ident);
                            RETURN_CODE RT = IDT_Device.SharedController.emv_callbackResponseLCD(EMV_LCD_DISPLAY_MODE.EMV_LCD_DISPLAY_MODE_AR_MESSAGE4, 0);

                        }
                    

                    }



                    break;
            }

        }


        private void displayCardData(IDTTransactionData cardData, string ident)
        {
            string text = "";
            if (cardData != null && cardData.Base64 != null && cardData.Base64.Length > 0)
                text += "Raw Data (Base 64):\r\n" + cardData.Base64 + "\r\n\r\n";
            if (cardData != null && cardData.fastEMV != null && cardData.fastEMV.Length > 0)
                text += "Fast EMV Data: " + cardData.fastEMV + "\r\n\r\n";
            if (cardData.Event == EVENT_TRANSACTION_DATA_Types.EVENT_TRANSACTION_PIN_DATA)
            {
                SetOutputText("PIN Data received:\r\n", ident);
                if (cardData.pin_KSN != null && cardData.pin_KSN.Length > 0)
                {
                    SetOutputText("KSN: " + cardData.pin_KSN + "\r\n", ident);

                }

                if (cardData.pin_pinblock != null && cardData.pin_pinblock.Length > 0)
                {
                    SetOutputText("PINBLOCK: " + cardData.pin_pinblock + "\r\n", ident);

                }

                if (cardData.pin_KeyEntry != null && cardData.pin_KeyEntry.Length > 0)
                {
                    SetOutputText("Key Entry: " + cardData.pin_KeyEntry + "\r\n", ident);

                }

                return;

            }

            if (cardData.Event == EVENT_TRANSACTION_DATA_Types.EVENT_TRANSACTION_DATA_CARD_DATA)
            {
                if (cardData.msr_rawData != null)
                {

                    SetOutputText("Raw Data (HEX): \r\n" + Common.getHexStringFromBytes(cardData.msr_rawData) + "\r\n", ident);

                }

            }

            if (cardData.captured_KSN != null)
            {
                //encrypted captured data
                if (cardData.captured_InitialVector != null)
                    text += "Man. Captured Initial Vector: " +
                            Common.getHexStringFromBytes(cardData.captured_InitialVector) + "\r\n";
                if (cardData.captured_KSN != null)
                    text += "Man. Captured KSN: " + Common.getHexStringFromBytes(cardData.captured_KSN) + "\r\n";
                if (cardData.captured_PAN != null)
                    text += "Man. Captured Encrypted PAN: " + Common.getHexStringFromBytes(cardData.captured_PAN) +
                            "\r\n";
                if (cardData.captured_Expiry != null)
                    text += "Man. Captured Encrypted Expiry: " +
                            Common.getHexStringFromBytes(cardData.captured_Expiry) + "\r\n";
                if (cardData.captured_CSC != null)
                    text += "Man. Captured Encrypted CSC: " + Common.getHexStringFromBytes(cardData.captured_CSC) +
                            "\r\n";
                if (cardData.captured_firstPANDigits != null && cardData.captured_firstPANDigits.Length > 0)
                    text += "PAN First Digits: " + cardData.captured_firstPANDigits + "\r\n";
                if (cardData.captured_lastPANDigits != null && cardData.captured_firstPANDigits.Length > 0)
                    text += "PAN Last Digits: " + cardData.captured_lastPANDigits + "\r\n";
                if (cardData.captured_MACValue != null)
                    text += "Man. Captured MAC: " + Common.getHexStringFromBytes(cardData.captured_MACValue) + "\r\n";
                if (cardData.captured_MACKSN != null)
                    text += "Man. Captured MAC KSN: " + Common.getHexStringFromBytes(cardData.captured_MACKSN) + "\r\n";




            }
            else if (cardData.captured_PAN != null)
            {
                //unecnrypted captured data
                if (cardData.captured_PAN != null)
                    text += "Man. Captured  PAN: " + System.Text.Encoding.UTF8.GetString(cardData.captured_PAN) +
                            "\r\n";
                if (cardData.captured_Expiry != null)
                    text += "Man. Captured  Expiry: " + System.Text.Encoding.UTF8.GetString(cardData.captured_Expiry) +
                            "\r\n";
                if (cardData.captured_CSC != null)
                    text += "Man. Captured  CSC: " + System.Text.Encoding.UTF8.GetString(cardData.captured_CSC) +
                            "\r\n";
            }



            if (cardData.device_RSN != null && cardData.device_RSN.Length > 0)
                text += "Serial Number: " + cardData.device_RSN + "\r\n";

            if (cardData.msr_track1Length > 0)
                text += "Track 1: " + cardData.msr_track1 + "\r\n";

            if (cardData.msr_encTrack1 != null)
                text += "Track 1 Encrypted: " + Common.getHexStringFromBytes(cardData.msr_encTrack1) + "\r\n";



            if (cardData.msr_hashTrack1 != null)
                text += "Track 1 Hash: " + Common.getHexStringFromBytes(cardData.msr_hashTrack1) + "\r\n";




            if (cardData.msr_track2Length > 0)
                text += "Track 2: " + cardData.msr_track2 + "\r\n";


            if (cardData.msr_encTrack2 != null)
                text += "Track 2 Encrypted: " + Common.getHexStringFromBytes(cardData.msr_encTrack2) + "\r\n";


            if (cardData.msr_hashTrack2 != null)
                text += "Track 2 Hash: " + Common.getHexStringFromBytes(cardData.msr_hashTrack2) + "\r\n";




            if (cardData.msr_track3Length > 0)
                text += "Track 3: " + cardData.msr_track3 + "\r\n";


            if (cardData.msr_encTrack3 != null)
                text += "Track 3 Encrypted: " + Common.getHexStringFromBytes(cardData.msr_encTrack3) + "\r\n";


            if (cardData.msr_hashTrack3 != null)
                text += "Track 3 Hash: " + Common.getHexStringFromBytes(cardData.msr_hashTrack3) + "\r\n";

            if (cardData.msr_KSN != null)
                text += "KSN: " + Common.getHexStringFromBytes(cardData.msr_KSN) + "\r\n";


            if (cardData.emv_clearingRecord != null)
            {
                if (cardData.emv_clearingRecord.Length > 0)
                {
                    text += "\r\nCTLS Clearing Record: \r\n";
                    text += Common.getHexStringFromBytes(cardData.emv_clearingRecord) + "\r\n";
                    Dictionary<string, string> dict = Common.processTLVUnencrypted(cardData.emv_clearingRecord);
                    foreach (KeyValuePair<string, string> kvp in dict) text += kvp.Key + ": " + kvp.Value + "\r\n";
                    text += "\r\n\r\n";

                }

            }

            if (cardData.emv_unencryptedTags != null)
            {
                if (cardData.emv_unencryptedTags.Length > 0)
                {
                    text += "\r\n======================== \r\n";

                    text += "\r\nUnencrypted Tags: \r\n";
                    text += Common.getHexStringFromBytes(cardData.emv_unencryptedTags) + "\r\n\r\n";
                    text += tlvToValues(cardData.emv_unencryptedTags);
                    text += "\r\n======================== \r\n";
                }

            }

            if (cardData.emv_encryptedTags != null)
            {
                if (cardData.emv_encryptedTags.Length > 0)
                {
                    text += "\r\n======================== \r\n";
                    text += "\r\nEncrypted Tags: \r\n";
                    text += Common.getHexStringFromBytes(cardData.emv_encryptedTags) + "\r\n\r\n";
                    text += tlvToValues(cardData.emv_encryptedTags);
                    text += "\r\n======================== \r\n";

                }

            }

            if (cardData.emv_maskedTags != null)
            {
                if (cardData.emv_maskedTags.Length > 0)
                {
                    text += "\r\n======================== \r\n";
                    text += "\r\nMasked Tags: \r\n";
                    text += Common.getHexStringFromBytes(cardData.emv_maskedTags) + "\r\n\r\n";
                    text += tlvToValues(cardData.emv_maskedTags);
                    text += "\r\n======================== \r\n";

                }

            }

            if (cardData.emv_hasAdvise) text += "CARD RESPONSE HAS ADVISE" + "\r\n";
            if (cardData.emv_hasReversal) text += "CARD RESPONSE HAS REVERSAL" + "\r\n";
            if (cardData.iccPresent == 1) text += "ICC Present: TRUE" + "\r\n";
            if (cardData.iccPresent == 2) text += "ICC Present: FALSE" + "\r\n";
            if (cardData.isCTLS == 1) text += "CTLS Capture: TRUE" + "\r\n";
            if (cardData.isCTLS == 2) text += "CTLS Capture: FALSE" + "\r\n";


            if (cardData.msr_extendedField != null && cardData.msr_extendedField.Length > 0)
                text += "Extended Field Bytes: " + Common.getHexStringFromBytes(cardData.msr_extendedField) + "\r\n";
      

            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_TRANS_ARMOR_PKI)
                text += "Encryption Type: TRANS_ARMOR_PKI\r\n";
            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_TRANS_ARMOR_TDES)
                text += "Encryption Type: TRANS_ARMOR_TDES\r\n";
            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_VOLTAGE)
                text += "Encryption Type: VOLTAGE\r\n";
            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_VISA_FPE)
                text += "Encryption Type: VISA_FPE\r\n";
            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_VERIFONE_FPE)
                text += "Encryption Type: VERIFONE_FPE\r\n";
            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_DESJARDIN)
                text += "Encryption Type: DESJARDIN\r\n";

            if (cardData.captureEncryptType == CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_TYPE_NONE)
                text += "Encryption Type: NONE\r\n";

            if (cardData.captureEncryptType != CAPTURE_ENCRYPT_TYPE.CAPTURE_ENCRYPT_TYPE_NONE)
            {
                if (cardData.msr_keyVariantType == KEY_VARIANT_TYPE.KEY_VARIANT_TYPE_DATA)
                    text += "Key Type: Data Variant\r\n";
                else if (cardData.msr_keyVariantType == KEY_VARIANT_TYPE.KEY_VARIANT_TYPE_PIN)
                    text += "Key Type: PIN Variant\r\n";
            }



            if (cardData.mac != null)
                text += "MAC: " + Common.getHexStringFromBytes(cardData.mac) + "\r\n";

            if (cardData.macKSN != null)
                text += "MAC KSN: " + Common.getHexStringFromBytes(cardData.macKSN) + "\r\n";

            if (cardData.msr_sessionID != null)
                text += "Session ID: " + Common.getHexStringFromBytes(cardData.msr_sessionID) + "\r\n";

            if (cardData.Event == EVENT_TRANSACTION_DATA_Types.EVENT_TRANSACTION_DATA_EMV_DATA)
            {
                switch (cardData.emv_resultCode)
                {
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_APPROVED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_APPROVED" + "\r\n");
                        SetOutputText("APPROVED", ident);
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_APPROVED_OFFLINE:
                        SetOutputText("APPROVED", ident);
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_APPROVED_OFFLINE" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_DECLINED_OFFLINE:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_DECLINED_OFFLINE" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_DECLINED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_DECLINED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_GO_ONLINE:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_GO_ONLINE" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_CALL_YOUR_BANK:
                        SetOutputText("CALL YOUR BANK", ident);
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_CALL_YOUR_BANK" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_NOT_ACCEPTED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_NOT_ACCEPTED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_FALLBACK_TO_MSR:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_FALLBACK_TO_MSR" + "\r\n");
                        SetOutputText("", ident);
                        SetOutputText(getLCDDisplayMessage(new byte[] { 0x45, 0x4E }, (int)DisplayMessages.DISPLAY_MESSAGE_IDENTIFIER.MSG_NEW_WELCOME), ident);
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_TIMEOUT:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_TIMEOUT" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_AUTHENTICATE_TRANSACTION:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_AUTHENTICATE_TRANSACTION" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_SWIPE_NON_ICC:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_SWIPE_NON_ICC" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_CTLS_TWO_CARDS:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_CTLS_TWO_CARDS" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_CTLS_TERMINATE:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_CTLS_TERMINATE" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_CTLS_TERMINATE_TRY_ANOTHER:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_CTLS_TERMINATE_TRY_ANOTHER" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_GO_ONLINE_CTLS:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_GO_ONLINE_CTLS" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_FAILED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_FAILED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_MSR_SWIPE_CAPTURED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_MSR_SWIPE_CAPTURED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_REQUEST_ONLINE_PIN:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_REQUEST_ONLINE_PIN" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_REQUEST_SIGNATURE:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_REQUEST_SIGNATURE" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_ADVISE_REQUIRED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_ADVISE_REQUIRED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_REVERSAL_REQUIRED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_REVERSAL_REQUIRED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_ADVISE_REVERSAL_REQUIRED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_ADVISE_REVERSAL_REQUIRED" + "\r\n");
                        break;
                    case EMV_RESULT_CODE.EMV_RESULT_CODE_NO_ADVISE_REVERSAL_REQUIRED:
                        text += ("EMV RESULT: " + "EMV_RESULT_CODE_NO_ADVISE_REVERSAL_REQUIRED" + "\r\n");
                        break;
                    default:
                        text += ("EMV RESULT: " + errorCode.getErrorString((RETURN_CODE)cardData.emv_resultCode) +
                                 "\r\n");
                        break;
                }
            }

            if (cardData.emv_transaction_Error_Code > 0)
                text += ("Transaction Error: " + errorCode.getTransError(cardData.emv_transaction_Error_Code) + "\r\n");
            if (cardData.emv_RF_State > 0)
                text += ("RF State: " + errorCode.getRFState(cardData.emv_RF_State) + "\r\n");
            if (cardData.emv_ESC > 0)
                text += ("Extended Status Code: " + errorCode.getExtendedStatusCode(cardData.emv_ESC) + "\r\n");
            if (cardData.emv_appErrorFn > 0)
                text += ("App Error Function: " + errorCode.getEMVAppErrorFn(cardData.emv_appErrorFn) + "\r\n");
            if (cardData.emv_appErrorState > 0)
                text += ("App Error State: " + errorCode.getEMVAppErrorState(cardData.emv_appErrorState) + "\r\n");



            if (cardData.ctlsApplication > 0)
            {
                text += "Contactless Application: ";
                switch (cardData.ctlsApplication)
                {
                    case CTLS_APPLICATION.CTLS_APPLICATION_AMEX:
                        text += ("CTLS_APPLICATION_AMEX" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_DISCOVER:
                        text += ("CTLS_APPLICATION_DISCOVER" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_MASTERCARD:
                        text += ("CTLS_APPLICATION_MASTERCARD" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_VISA:
                        text += ("CTLS_APPLICATION_VISA" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_SPEEDPASS:
                        text += ("CTLS_APPLICATION_SPEEDPASS" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_GIFT_CARD:
                        text += ("CTLS_APPLICATION_GIFT_CARD" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_DINERS_CLUB:
                        text += ("CTLS_APPLICATION_DINERS_CLUB" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_EN_ROUTE:
                        text += ("CTLS_APPLICATION_EN_ROUTE" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_JCB:
                        text += ("CTLS_APPLICATION_JCB" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_VIVO_DIAGNOSTIC:
                        text += ("CTLS_APPLICATION_VIVO_DIAGNOSTIC" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_HID:
                        text += ("CTLS_APPLICATION_HID" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_MSR_SWIPE:
                        text += ("CTLS_APPLICATION_MSR_SWIPE" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_RESERVED:
                        text += ("CTLS_APPLICATION_RESERVED" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_DES_FIRE_TRACK_DATA:
                        text += ("CTLS_APPLICATION_DES_FIRE_TRACK_DATA" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_DES_FIRE_RAW_DATA:
                        text += ("CTLS_APPLICATION_DES_FIRE_RAW_DATA" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_RBS:
                        text += ("CTLS_APPLICATION_RBS" + "\r\n");
                        break;
                    case CTLS_APPLICATION.CTLS_APPLICATION_VIVO_COMM:
                        text += ("CTLS_APPLICATION_VIVO_COMM" + "\r\n");
                        break;

                }
            }



            SetOutputText(text, ident);

            text = "";
            byte[] temp = new byte[0];

   


        }

        private string tlvToValues(byte[] tlv)
        {
            string text = "";
            Dictionary<string, string> dict = Common.processTLVUnencrypted(tlv);
            foreach (KeyValuePair<string, string> kvp in dict) text += kvp.Key + ": " + kvp.Value + "\r\n";
            dict.Clear();
            dict = null;
            return text;
        }


        private void showMessageOnLCD(byte[] language, byte[] message, string ident)
        {
            if (message == null) return;
            StringBuilder sb = new StringBuilder();
            int position = 0;
            while (position < message.Length)
            {
                byte line = message[position++];
                if (position < message.Length)
                {
                    if ((line & 0x80) == 0x80)
                    {
                        sb.Append(getLCDDisplayMessage(language, message[position++]));

                    }
                    else
                    {
                        byte[] msgs = new byte[message.Length - 2];
                        Array.Copy(message, 1, msgs, 0, msgs.Length);
                        label4.Text = ASCIIEncoding.ASCII.GetString(msgs);
                        SetOutputText("\r\n" + ASCIIEncoding.ASCII.GetString(msgs) + "\r\n", ident);
                        return;
                    }

                    

                }


            }

            label4.Text = sb.ToString();

            System.Diagnostics.Debug.WriteLine("Setting LCD Screen Message");
            SetOutputText("\r\n" + sb.ToString() + "\r\n", ident);
            

        }

        public static string getLCDDisplayMessage(byte[] language, int id)
        {
            if (id > (int)DisplayMessages.DISPLAY_MESSAGE_IDENTIFIER.TRANS_MSG_NEW_TBL_LIMIT) return "";
            ASCIIEncoding encoding = new ASCIIEncoding();
            string line1 = "";
            string line2 = "";
            string sLanguage = encoding.GetString(language, 0, 2).ToUpper();
            DisplayMessages.DISPLAY_MESSAGE_LANGUAGE theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.ENGLISH;
            if (sLanguage.Equals("FR"))
            {
                theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.FRENCH;
            }
            if (sLanguage.Equals("EF"))
            {
                theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.ENGLISH_FRENCH;
            }
            if (sLanguage.Equals("ES"))
            {
                theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.SPANISH;
            }
            if (sLanguage.Equals("PT"))
            {
                theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.PORTUGUESE;
            }
            if (sLanguage.Equals("ZH"))
            {
                theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.CHINESE;
            }
            if (sLanguage.Equals("JA"))
            {
                theLang = DisplayMessages.DISPLAY_MESSAGE_LANGUAGE.JAPANESE;
            }

            DisplayMessages.retrieveMessage(theLang, (DisplayMessages.DISPLAY_MESSAGE_IDENTIFIER)id, ref line1, ref line2);
            return (line1 + "\r\n" + line2);
        }


        static object tabLock = new object();

        Dictionary<string, TextBox> resultsDict = new Dictionary<string, TextBox>();
        Dictionary<string, Button> buttonDict = new Dictionary<string, Button>();

        private object getTabObject(string ident, int type)
        {
            //type 0 = message, type 1 = result, type 2 = log

  
            if (type == 1)
            {
                if (resultsDict.ContainsKey(ident)) return resultsDict[ident];
            }
 
            if (type == 3)
            {
                if (buttonDict.ContainsKey(ident)) return buttonDict[ident];
            }


            TabPage page = getLogTab(ident);
            if (page == null) return null;
                foreach (Control tb in page.Controls)
                {
                    string tag = (string)tb.Tag;
                if (tag != null && tag.Length > 0)
                {
    
                    if (type == 1 && tag.Equals("result"))
                    {
                        if (resultsDict.ContainsKey(ident)) resultsDict.Remove(ident);
                        resultsDict.Add(ident, (TextBox)tb);
                        return (TextBox)tb;
                    }
       
                    if (type == 3 && tag.Equals(ident))
                    {
                        if (buttonDict.ContainsKey(ident)) buttonDict.Remove(ident);
                        buttonDict.Add(ident, (Button)tb);
                        return (Button)tb;
                    }
                }

                }

                return null;
            

        }

        private object getTabObject(TabPage page, int type)
        {
            if (page == null) return null;
            //type 0 = message, type 1 = result, type 2 = log
            string ident = page.Text;
      
            if (type == 1)
            {
                if (resultsDict.ContainsKey(ident)) return resultsDict[ident];
            }

            if (type == 3)
            {
                if (buttonDict.ContainsKey(ident)) return buttonDict[ident];
            }


            foreach (Control tb in page.Controls)
            {
                string tag = (string)tb.Tag;
                if (tag != null && tag.Length > 0)
                {
           
                    if (type == 1 && tag.Equals("result"))
                    {
                        if (resultsDict.ContainsKey(ident)) resultsDict.Remove(ident);
                        resultsDict.Add(ident, (TextBox)tb);
                        return (TextBox)tb;
                    }

                    if (type == 3 && tag.Equals(ident))
                    {
                        if (buttonDict.ContainsKey(ident)) buttonDict.Remove(ident);
                        buttonDict.Add(ident, (Button)tb);
                        return (Button)tb;
                    }
                }

            }

            return null;


        }

        private void removeTab(string ident)
        {
            lock (tabLock)
            {
                try
                {

                    if (Devices != null && Devices.TabCount == 1 && Devices.TabPages[0].Text.Equals(ident))
                    {
                        
                        TextBox tb = (TextBox)getTabObject(Devices.TabPages[0], 1);
                        if (tb == null) return;
                        tb.Text = "";

                        Button b = (Button)getTabObject(Devices.TabPages[0], 3);
                        Devices.TabPages[0].Text = "No Device";
                        if (b == null) return;
                        b.Tag = "No Device";
                        return;
                    }

                    TabPage foundPage = null;
                    foreach (TabPage tabpage in Devices.TabPages)
                    {
                        if (tabpage.Text.Equals(ident))
                        {
                            foundPage = tabpage;
                            break; ;
                        }
                    }
                    if (foundPage != null) Devices.TabPages.Remove(foundPage);
                }
                catch
                {

                }

            }
            
        }


     
        string[] properties = new string[] { "BackColor", "Dock", "Font", "Location", "Margin", "Multiline", "Size", "Tag", "Text", "TextAlign", "ReadOnly", "ScrollBars", "AutoSize", "Padding", "UseVisualStyleBackColor" };

        private bool copySource(string name)
        {
            if (name == null || name.Length == 0) return false;
            foreach (string prop in properties)
            {
                if (name.Equals(prop)) return true;
            }
            return false;
        }

        private TabPage getLogTab(string ident)
        {
            lock (tabLock)
            {
                if (ident == null) return null;
                if (ident.Equals("app")) return null;
                if (ident.Length == 0) return null;
                if (Devices.TabCount == 1 && Devices.TabPages[0].Text.Equals("No Device"))
                {

                    TextBox tb = (TextBox)getTabObject(Devices.TabPages[0], 1);
                    tb.Text = "";

                    Button b = (Button)getTabObject(Devices.TabPages[0], 3);
                    Devices.TabPages[0].Text = ident;
                    b.Tag = ident;
                    if (!config.ContainsKey(ident))
                    {
                        config.Add(ident, new Config());
                    }
                    return Devices.TabPages[0];
                }


                foreach (TabPage tabpage in Devices.TabPages)
                {
                    if (tabpage.Text.Equals(ident)) return tabpage;
                }
                //doesn't exist, so let's create a new output window
                TabPage page = new TabPage();


                foreach (PropertyInfo sourceProperty in Devices.TabPages[0].GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance | BindingFlags.SetProperty))
                {
                    if (copySource(sourceProperty.Name))
                    {
                        object newValue = sourceProperty.GetValue(Devices.TabPages[0], null);
                        sourceProperty.SetValue(page, newValue, null);

                    }
                }


                page.Text = ident;


                Button clear = new Button();
                page.Controls.Add(clear);

                Label label582 = new Label();
                page.Controls.Add(label582);
                TextBox tbOutput2 = new TextBox();
                page.Controls.Add(tbOutput2);


       

                if (resultsDict.ContainsKey(ident)) resultsDict.Remove(ident);
                resultsDict.Add(ident, (TextBox)tbOutput2);

                if (buttonDict.ContainsKey(ident)) buttonDict.Remove(ident);
                buttonDict.Add(ident, (Button)clear);

                foreach (PropertyInfo sourceProperty in ClearLogsButton.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance | BindingFlags.SetProperty))
                {
                    if (copySource(sourceProperty.Name))
                    {
                        object newValue = sourceProperty.GetValue(ClearLogsButton, null);
                        sourceProperty.SetValue(clear, newValue, null);
                    }

                }

                clear.Tag = ident;
                clear.Click += new System.EventHandler(this.btnClearLogs_Click_1);

 

                foreach (PropertyInfo sourceProperty in label58.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance | BindingFlags.SetProperty))
                {
                    if (copySource(sourceProperty.Name))
                    {
                        object newValue = sourceProperty.GetValue(label58, null);
                        sourceProperty.SetValue(label582, newValue, null);
                    }
                }

                foreach (PropertyInfo sourceProperty in tbOutput.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance | BindingFlags.SetProperty))
                {
                    if (copySource(sourceProperty.Name))
                    {
                        object newValue = sourceProperty.GetValue(tbOutput, null);
                        sourceProperty.SetValue(tbOutput2, newValue, null);
                    }
                }

      
                if (!config.ContainsKey(ident))
                {
                    config.Add(ident, new Config());
                }
                tbOutput2.Text = "";
                Devices.Controls.Add(page);

                clear.BringToFront();
                label582.BringToFront();
                tbOutput2.BringToFront();
         

                return page;
            }
           
        }


        private void enableButton(string ident)
        {
            if (InvokeRequired)
            {
                Invoke((MethodInvoker)delegate { enableButton(ident); });
                return;
            }
            TabPage tp = getLogTab(ident);
            if (tp != null)
            {
                Button b = (Button)getTabObject(Devices.TabPages[0], 3);
                if (b != null) b.Enabled = true;
            }
        }

        private void disableButton(string ident)
        {
            if (InvokeRequired)
            {
                Invoke((MethodInvoker)delegate { disableButton( ident); });
                return;
            }
            TabPage tp = getLogTab(ident);
            if (tp != null)
            {
                Button b = (Button)getTabObject(Devices.TabPages[0], 3);
                if (b != null) b.Enabled = false;
            }
        }


        void WPCallback(IDT_DEVICE_Types sender, string xmlResponse, RETURN_CODE resultCode, EMV_RESULT_CODE emvCode, string ident, IDTTransactionData cardData, WorldPay.ReceiptData receipt)
        {
            if (InvokeRequired)
            {
                Invoke((MethodInvoker)delegate { WPCallback( sender,  xmlResponse,  resultCode,  emvCode,  ident,  cardData, receipt); });
                return;
            }
            if (cardData != null)
            {
                SetOutputText("\r\n\r\n");
                displayCardData(cardData, ident);
            }
            SetOutputText("\r\n\r\nResponse received from Worldpay: " + resultCode.ToString() + "\r\n\r\n", ident);
            SetOutputText(xmlResponse, ident);
            if (cardData != null)
            {
                SetOutputText(tsBox.Text + "\r\n\r\n", ident);
                //SetOutputText(ReceiptCTLS(cardData, xmlResponse), ident);
            }
            if (receipt != null) SetOutputText(getReceipt(receipt), ident);

        }


        public int dialogResponse = 0;
        private void btnClearLogs_Click_1(object sender, EventArgs e)
        {
            WorldPay.bypassNetworkChecking = bypassNetworkChecking.Checked;
            IDT_Device.setCallbackIP(MessageCallBack);

            Button b = (Button)sender;


            
            

            string ident = (string)b.Tag;
            SetOutputText(null, ident);
            WorldPay wp = new WorldPay();


            if (forwaredTransaction.Checked)
            {

                try
                {
                    string pw = "";
                    if (usePassword.Checked) pw = tbPassword.Text;
                    RETURN_CODE fwd = wp.forwardTransaction(WPCallback, forwardID.Text, pw, ident,cbDelete.Checked);
                    if (fwd == RETURN_CODE.RETURN_CODE_DO_SUCCESS)
                    {
                        if (cbDelete.Checked)
                        {
                            tbOutputAppendText("Transaction successfully deleted from stored transactions\r\n", IDT_Device.currentIdent());

                        }
                        else
                        {
                            tbOutputAppendText("Forward Transaction Request Sent Successful\r\n", IDT_Device.currentIdent());

                        }
                    }
                    else
                    {
                        if (cbDelete.Checked)
                        {
                            tbOutputAppendText("Delete Stored Transaction failed Error Code: " + "0x" + String.Format("{0:X}", (ushort)fwd) +
                    ": " + IDTechSDK.errorCode.getErrorString(fwd) + "\r\n", IDT_Device.currentIdent());
                        }
                        else
                        {
                            tbOutputAppendText("Forward Transaction Request Sent failed Error Code: " + "0x" + String.Format("{0:X}", (ushort)fwd) +
                                                ": " + IDTechSDK.errorCode.getErrorString(fwd) + "\r\n", IDT_Device.currentIdent());
                        }


                    }
                    return;
                }
                catch
                {
                    SetOutputText("Invalid Stored Transaction GUID provided");
                    return;
                }


            }






            WorldPay.WorldPayData data = new WorldPay.WorldPayData();

            data.storeTransaction = storeTransaction.Checked;
            if (usePassword.Checked) data.storagePassword = tbPassword.Text;
            if (cbMemo.Checked) data.storageMemo = tbMemo.Text;
            data.declinePartial = declinePartial.Checked;
            data.transactionID = tbTransactionID.Text;
            data.deviceID = ident;
            data.accountID = accountID.Text;
            data.license = license.Text;
            data.storeOnHostNetworkError = cbStore.Checked;
            data.accountToken = accountToken.Text;
            data.acceptorID = acceptorID.Text;
            data.transactionType = (WorldPay.TRANSACTION_TYPE)transactiontype.SelectedIndex;
            data.amount = Amount.Text;
            data.timeout = Convert.ToInt32(timeout.Text);
            data.isTest = demo.Checked;
            data.msrOnly = msronly.Checked;
            data.referenceNumber = referencenumber.Text;
            data.terminalID = terminalid.Text;
            data.ticketNumber = ticketnumber.Text;
            data.duplicateCheck = duplicatecheckdisableflag.Checked;
            data.duplicateOverride = duplicateoverrideflag.Checked;
            data.enableCTLS = enableCTLS.Checked;
            data.isSimulated = simulated.Checked;
            wp.returnRequest = returnRequest.Checked;
            data.interfaceOverride = interfaceOverride.SelectedIndex;
            data.authAmount = authAmount.Text;
            data.additionalTags = addTags.Text;

            RETURN_CODE rt = wp.executeTransaction(data, WPCallback,false);

            if (rt == RETURN_CODE.RETURN_CODE_DO_SUCCESS)
            {
                tbOutputAppendText("\r\n\r\n\r\nStart Transaction Successful\r\n", IDT_Device.currentIdent());
            }
            else
            {
                tbOutputAppendText("Start transaction failed Error Code: " + "0x" + String.Format("{0:X}", (ushort)rt) +
                                    ": " + IDTechSDK.errorCode.getErrorString(rt) + "\r\n", IDT_Device.currentIdent());

            }

        }

        private string getIdentValue()
        {
            string str = IDT_Device.currentIdent();
            return str;


        }


        private string sel = "";
        private void connectedDevicesList_SelectedIndexChanged(object sender, EventArgs e)
        {
            sel = (string)connectedDevicesList.SelectedItem;
            if (sel != null)
            {
                TabPage pg = getLogTab(sel);
                if (pg != null) Devices.SelectedTab = pg;
                IDT_Device.setDeviceType(sel);
                
            }
            else sel = "";


        }

        bool formIsClosing = false;
        private void UniversalSDKDemo_FormClosed(object sender, FormClosedEventArgs e)
        {
            formIsClosing = true;
            Application.Exit();
        }


    
        //Symmetric RKI B
   


        private void tbOutputAppendText(string text, string ident)
        {
            // InvokeRequired required compares the thread ID of the
            // calling thread to the thread ID of the creating thread.
            // If these threads are different, it returns true.
            if (tbOutput.InvokeRequired)
            {
                try
                {
                    SetTextCallback d = new SetTextCallback(tbOutputAppendText);
                    Invoke(d, new object[] { text, ident });
                }
                catch
                {

                }

            }
            else
            {
                if (ident == null || ident.Length == 0) ident = IDT_Device._path;
                TextBox tb = (TextBox)getTabObject(ident, 1);
                if (tb == null) tb = sdkOutput;
                if (text == null || text.Length == 0)
                {
                    tb.Text = "";
                    return;
                }
                try
                {
                    tb.AppendText(text);
                }
                catch (Exception)
                {
                }

            }

            text = "";
        }


        private void SetOutputText(string text, string ident = "app")
        {
            // InvokeRequired required compares the thread ID of the
            // calling thread to the thread ID of the creating thread.
            // If these threads are different, it returns true.
            if (tbOutput.InvokeRequired)
            {
                try
                {
                    SetTextCallback d = new SetTextCallback(SetOutputText);
                    Invoke(d, new object[] { text, ident });
                }
                catch
                {

                }

            }
            else
            {
                if (ident == null || ident.Length == 0) ident = IDT_Device._path;
                TextBox tb = null;
                if (ident.Equals("app")) tb = sdkOutput;
                else tb = (TextBox)getTabObject(ident, 1);
                if (tb == null) tb = sdkOutput;
                if (text == null || text.Length == 0)
                {
                    tb.Text = "";
                    return;
                }
                try
                {
                    text = text.TrimEnd('\r', '\n');
                    tb.AppendText(text + "\r\n");
                }
                catch (Exception)
                {
                }

            }

            text = "";
        }


        private void SetTitleText(string text)
        {
            // InvokeRequired required compares the thread ID of the
            // calling thread to the thread ID of the creating thread.
            // If these threads are different, it returns true.
            if (this.InvokeRequired)
            {
                SetTitleTextCallback d = new SetTitleTextCallback(SetTitleText);
                Invoke(d, new object[] { text });
            }
            else
            {
                this.Text = text;
            }
        }


      
        private void rescanForDevicesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Profile.discoverDevice();
        }

        private int baud = 0;
        private List<int> getavailablePorts()
        {
            List<int> ports = new List<int>();

            string[] std = SerialPort.GetPortNames();
            if (std != null && std.Length > 0)
            {
                foreach (string str in std)
                {
                    if (str != null && str.StartsWith("COM"))
                    {

                        int port = 0;
                        try
                        {
                            port = Convert.ToInt32(str.Substring(3));
                            if (connectedDevicesList.Items.Count > 0)
                            {
                                foreach (string device in connectedDevicesList.Items)
                                {
                                    if (device.EndsWith(str)) port = 0;

                                }
                            }
                            if (port > 0) ports.Add(port);
                        }
                        catch
                        {

                        }

                    }
                }

            }

            return ports;
        }

        private void bTMagToolStripMenuItem_Click(object sender, EventArgs e)
        {
            List<int> availablePorts = getavailablePorts();
            if (availablePorts == null || (availablePorts != null && availablePorts.Count == 0))
            {
                MessageBox.Show("No available COM ports to open", "No COM Ports", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            ToolStripMenuItem item = (ToolStripMenuItem)sender;


            IDT_DEVICE_Types type = Profile.deviceTypeFromName(item.Text.Trim());
            int port = availablePorts[0];

            if (availablePorts.Count > 1)
            {
                StringBuilder sb = new StringBuilder();
                sb.AppendLine("Please choose from the following COM ports:");
                foreach (int val in availablePorts)
                {
                    sb.Append(val.ToString() + "   ");
                }
                sb.AppendLine("");
                string input2 =
                                Microsoft.VisualBasic.Interaction.InputBox(sb.ToString(), "Data",
                                    availablePorts[0].ToString(), -1, -1);

                int selectedPort = 0;
                try
                {
                    selectedPort = Convert.ToInt32(input2);
                }
                catch
                {

                }
                if (selectedPort > 0) port = selectedPort;
            }


            bool isConnected = false;
            if (baud == 0)
            {
                isConnected = IDT_Device.useSerialPort(type,
                     port);
            }
            else
            {
                isConnected = IDT_Device.useSerialPort(type,
                    port, baud);
            }


            if (!isConnected)
            {

                tbOutputAppendText("Failed to locate device on specified COM port", getIdentValue());
                return;
            }

            IDT_Device.setDeviceType(type, DEVICE_INTERFACE_Types.DEVICE_INTERFACE_SERIAL);
        }

        private void connectUsingTLS12ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            connectIP(true);
        }

        private void connectWithoutTLS12ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            connectIP(false);
        }

        private void connectIP(bool useTLS)
        {
            string ident = getIdentValue();



            string input = Microsoft.VisualBasic.Interaction.InputBox("What IP Address Is The VP6800 Located At?",
                "IP Address", _lastIP, -1, -1);

            if (input.Length == 0)
            {
                tbOutputAppendText("No IP Address specified.  \r\n", ident);
                return;
            }

            //IDT_Device.SharedController.ip_monitorSocketConnectionStatus(true, false, 10, 0);
            if (!IDT_Device.SharedController.ip_connectToSocket(input, useTLS))
            {
                tbOutputAppendText("Attempted connection to \"" + input + "\" failed.  \r\n", ident);
                return;
            }

            _lastIP = input;
            refreshDevices();
        }

        private void disconnectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (connectedDevicesList.SelectedItem != null)
            {
                Profile.closeDevice((string)connectedDevicesList.SelectedItem);
                refreshDevices();
            }
            else
            {
                MessageBox.Show("No device selected that can be disconnected.", "No Device", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void bDefault_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem item = (ToolStripMenuItem)sender;
            bDefault.Checked = false;
            b2400.Checked = false;
            b4800.Checked = false;
            b9600.Checked = false;
            b19200.Checked = false;
            b38400.Checked = false;
            b57600.Checked = false;
            b115200.Checked = false;
            item.Checked = true;
            if (item.Name.Equals("bDefault")) baud = 0;
            else baud = Convert.ToInt32(item.Text);
        }
        string currentIPAddress = "";
        private void standbyModeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string ident = getIdentValue();
            string address = currentIPAddress;
            if (currentIPAddress == null || currentIPAddress.Length == 0) address = "10.12.32.98";
            string input = Microsoft.VisualBasic.Interaction.InputBox(
                "What IP Address Is NEO 2 Device To Put Into Standby Located At?", "Close IP Address", address, -1, -1);

            if (input.Length == 0)
            {
                tbOutputAppendText("No IP Address specified.  \r\n", ident);

                return;
            }

            RETURN_CODE rt = IDT_Device.SharedController.device_enterStandbyMode(input);
            if (rt == RETURN_CODE.RETURN_CODE_DO_SUCCESS)
            {
                tbOutputAppendText("Standby Mode Successful\r\n", ident);
            }
            else
            {
                tbOutputAppendText("Standby Mode Error Code: " + "0x" + String.Format("{0:X}", (ushort)rt) + ": " +
                                    IDTechSDK.errorCode.getErrorString(rt) + "\r\n", ident);

            }
        }

        private void wakeAllDevicesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string ident = getIdentValue();
            string input =
                Microsoft.VisualBasic.Interaction.InputBox("What IP Address To Wake?", "IP Address", "10.12.34.42", -1,
                    -1);

            if (input.Length == 0)
            {
                tbOutputAppendText("Invalid IP address.  \r\n", ident);
                return;
            }



            RETURN_CODE rt = IDT_Device.device_wakeDevice("", input);
            if (rt == RETURN_CODE.RETURN_CODE_DO_SUCCESS)
            {
                tbOutputAppendText("Wake Executed Successful\r\n", ident);
            }
            else
            {
                tbOutputAppendText("Wake  Error Code: " + "0x" + String.Format("{0:X}", (ushort)rt) + ": " +
                                    IDTechSDK.errorCode.getErrorString(rt) + "\r\n", ident);

            }
        }

        private void wakeDeviceByMACToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string ident = getIdentValue();
            string input = Microsoft.VisualBasic.Interaction.InputBox("What MAC Address To Wake?", "MAC Address",
                "782079001028", -1, -1);

            if (input.Length != 12)
            {
                tbOutputAppendText("Invalid MAC address.  \r\n", ident);
                return;
            }

            RETURN_CODE rt = IDT_Device.device_wakeDevice(input);
            if (rt == RETURN_CODE.RETURN_CODE_DO_SUCCESS)
            {
                tbOutputAppendText("Wake Executed Successful\r\n", ident);
            }
            else
            {
                tbOutputAppendText("Wake  Error Code: " + "0x" + String.Format("{0:X}", (ushort)rt) + ": " +
                                    IDTechSDK.errorCode.getErrorString(rt) + "\r\n", ident);

            }
        }

        bool inLoad = false;
 
     

        private void saveSettings(object sender, EventArgs e)
        {
            if (inLoad) return;
           // Settings.Default.accountid = accountID.Text;
          //  Settings.Default.accounttoken = accountToken.Text;
          //  Settings.Default.acceptorid = acceptorID.Text;
          //  Settings.Default.terminalid = terminalid.Text;


           // Settings.Default.Save();
        }

        private void configureSelectedDeviceToolStripMenuItem_Click(object sender, EventArgs e)
        {

            if (MessageBox.Show("Start device configuration with WorldPay Terminal Settings, AIDs and CAPKs?", "Configure Device?", MessageBoxButtons.OKCancel, MessageBoxIcon.Question) != DialogResult.OK)
            {

                return;
            }
            string ident = (string)connectedDevicesList.SelectedItem;
            string jsonfile = "";
            try
            {
                jsonfile = System.IO.File.ReadAllText(Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location) + "\\Worldpay_Device_Config.json");
            }
            catch
            {
                tbOutputAppendText("Could not read configuration file from disk.\r\n", ident);
                return;
            }

            RETURN_CODE rt = IDT_Device.SharedController.device_sendConfiguration(jsonfile, VC_OPERATION_TYPE.VC_OPERATION_TYPE_WRITE_IGNORE_HASH, false, ident);
            ////RETURN_CODE rt = IDT_Device.SharedController.device_sendConfiguration(file, type, abortFirmware.Checked, ident);
            if (rt == RETURN_CODE.RETURN_CODE_DO_SUCCESS)
            {
                tbOutputAppendText("ViVOconfig Success: \r\n", ident);
            }
            else
            {
                tbOutputAppendText("ViVOconfig Result Code: " + "0x" + String.Format("{0:X}", (ushort)rt) +
                                    ": " + IDTechSDK.errorCode.getErrorString(rt) + "\r\n", ident);

            }
        }

        bool lastOnline = false;

        public string getReceipt(WorldPay.ReceiptData receipt)
        {
            if (receipt == null) return "";
            String rec = "";
            rec = rec + "================================";
            rec = rec + printCenter("\r\n");
            rec = rec + printCenter("RECEIPT");
            rec = rec + printCenter("\r\n");

            rec = rec + "================================";
            rec = rec + printCenter("\r\n");
            string dateString = receipt.dateTime.ToString("yyyy/MM/dd HHmmss");
            rec = rec + printCenter("Joe's Resaturant");
            rec = rec + printCenter("\r\n");

            rec = rec + printCenter("10721 Walker Street");
            rec = rec + printCenter("\r\n");
            rec = rec + printCenter("Cypress, CA 90630");
            rec = rec + printCenter("\r\n");
            rec = rec + printCenter("714-761-8886");
            rec = rec + printCenter("\r\n");
            rec = rec + printCenter("www.joesrestaurant.net");
            rec = rec + printCenter("\r\n");



            rec = rec + "--------------------------------";
            rec = rec + "\r\n";
            switch (receipt.saleType)
            {
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_SALE:
                    rec = rec + printCenter("SALE");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_PRE_AUTH:
                    rec = rec + printCenter("PRE-AUTH");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_COMPLETION:
                    rec = rec + printCenter("COMPLETION");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_VOID:
                    rec = rec + printCenter("VOID");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_RETURN:
                    rec = rec + printCenter("RETURN");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_REVERSAL_FULL:
                    rec = rec + printCenter("FULL REVERSAL");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_REVERSAL_PARTIAL:
                    rec = rec + printCenter("PARTIAL REVERSAL");
                    break;
                case WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_CREDIT:
                    rec = rec + printCenter("CREDIT");
                    break;
                default:
                    break;
            }
            rec = rec + "\r\n";
            if (string.IsNullOrEmpty(receipt.approvedAmount)) receipt.approvedAmount = "0.00";
            if (string.IsNullOrEmpty(receipt.result)) receipt.result = "Unknown Response";
            if (!receipt.result.ToUpper().Contains("PARTIAL"))
            {
                if (!receipt.isApproved)
                {
                    rec = rec + printCenter("TRANSACTION FAILED");
                }
                else
                {
                    rec = rec + printCenter("TRANSACTION SUCCESSFUL");
                }
                rec = rec + "\r\n";
                rec = rec + printCenter(receipt.result);
            }
            else
            {
                if (receipt.approvedAmount.Equals("0.00"))
                {
                    rec = rec + printCenter("TRANSACTION FAILED");
                    rec = rec + "\r\n";
                    rec = rec + printCenter("PARTIAL APPROVAL DECLINED");
                }
                else
                {
                    rec = rec + printCenter("PARTIAL AMOUNT APPROVED");
                }
            }

         

            rec = rec + "\r\n";

            if (receipt.entryMethod == WorldPay.ENTRY_METHOD.ENTRY_METHOD_SWIPE) rec = rec + printJustified("Entry Method: ", "SWIPE");
            if (receipt.entryMethod == WorldPay.ENTRY_METHOD.ENTRY_METHOD_CTLS) rec = rec + printJustified("Entry Method: ", "CTLS");
            if (receipt.entryMethod == WorldPay.ENTRY_METHOD.ENTRY_METHOD_CONTACT) rec = rec + printJustified("Entry Method: ", "CONTACT");

            rec = rec + "\r\n";
            rec = rec + "--------------------------------";
            
            if (!String.IsNullOrEmpty(receipt.card))
            {
                rec = rec + "\r\n";
                rec = rec + printJustified("Card:", receipt.card);
            }
            if (!String.IsNullOrEmpty(receipt.account))
            {
                rec = rec + "\r\n";
                rec = rec + printJustified("Account:", receipt.account);
            }
            if (!String.IsNullOrEmpty(receipt.transType))
            {
                rec = rec + "\r\n";
                rec = rec + printJustified("Transaction Type:", receipt.transType);
            }
            rec = rec + "\r\n";
            rec = rec + printJustified("Date/Time:", dateString);

                     
            if (!receipt.approvedAmount.Equals("0.00"))
            {
                rec = rec + "\r\n";
                rec = rec + printJustified("Approved Amount:", receipt.approvedAmount);
            }
            else if (receipt.saleType == WorldPay.TRANSACTION_TYPE.TRANSACTION_TYPE_RETURN && receipt.isApproved)
            {
                rec = rec + "\r\n";
                rec = rec + printJustified("Return Amount:", receipt.amount);
            }
            rec = rec + "\r\n";
            if (!String.IsNullOrEmpty(receipt.iad))
            {
                rec = rec + printJustified("IAD:", receipt.iad);
                rec = rec + "\r\n";
            }
            if (!String.IsNullOrEmpty(receipt.aid))
            {
                rec = rec + printJustified("AID:", receipt.aid);
                rec = rec + "\r\n";
            }
            if (!String.IsNullOrEmpty(receipt.authCode))
            {
                rec = rec + printJustified("AUTH:", receipt.authCode);
                rec = rec + "\r\n";
            }
            if (!String.IsNullOrEmpty(receipt.transID))
            {
                rec = rec + printJustified("TRANS ID:", receipt.transID);
                rec = rec + "\r\n";
            }
            if (!String.IsNullOrEmpty(receipt.cert))
            {
                rec = rec + printJustified("TC:", receipt.cert);
                rec = rec + "\r\n";
            }
            if (!String.IsNullOrEmpty(receipt.authReversal))
            {
                rec = rec + printJustified("Reversal Amt:", receipt.authReversal);
                rec = rec + "\r\n";
            }
            if (!String.IsNullOrEmpty(receipt.tokenSubstitute))
            {
                rec = rec + printJustified("Token Sub:", receipt.tokenSubstitute);
                rec = rec + "\r\n";
            }

            if (receipt.cvm == WorldPay.CVM_TYPE.CVM_TYPE_NO_CVM)
            {
                rec = rec + "\r\n\r\nNO CVM\r\n\r\n";

            }
            if (receipt.cvm == WorldPay.CVM_TYPE.CVM_TYPE_SIGNATURE)
            {
                if (MessageBox.Show("Would you like to bypass signature collection?",
                   "Bypass Signature", MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    rec = rec + "\r\n\r\nSIGNATURE BYPASS";
                }
                else
                {
                    rec = rec + "Card Holder Signature:\r\n\r\n\r\n\r\n------------------------------\r\n" + receipt.name + "\r\n\r\nI Acknowledge Satisfactory Rec-\r\n";
                    rec = rec + "eipt Of Relative Goods/Services\r\n\r\n";
                }

            }

            rec = rec + "\r\n\r\n";
            return rec;

        }

        private string printCenter(String str)
        {
            if (str == null) str = "";
            int len = str.Length;
            if (len > 30)
            {
                return str;
            }
            for (int x = 0; x < (32 - len) / 2; x++)
            {
                str = " " + str;
            }
            return str;
        }


        private string printJustified(String str, String str2)
        {
            if (str == null) str = "";
            if (str2 == null) str2 = "";
            int len = str.Length + str2.Length;
            if (len > 31)
            {
                return str + str2;
            }
            for (int x = len; x < 32; x++)
            {
                str2 = " " + str2;
            }
            return str + str2;
        }

        private void button2_Click(object sender, EventArgs e)
        {

            WorldPay wp = new WorldPay();
            
            if (wp.hostReachable())
            {
                SetOutputText("Worldpay Servers Can Be Reached");
            }
            else
            {
                SetOutputText("Worldpay Servers Cannot Be Reached");
            }
            
        }

        private void button1_Click(object sender, EventArgs e)
        {
            WorldPay wp = new WorldPay();
            string json = wp.getSAFListing();
            if (String.IsNullOrEmpty(json))
            {
                SetOutputText("No Stored Transactions Found");
            }
            else
            {
                SetOutputText(json);
            }

        }

        private void storeTransaction_CheckedChanged(object sender, EventArgs e)
        {
            if (storeTransaction.Checked) forwaredTransaction.Checked = false;
        }

        private void forwaredTransaction_CheckedChanged(object sender, EventArgs e)
        {
            if (forwaredTransaction.Checked) storeTransaction.Checked = false;
        }
    }

}
